require 'sqlite3'
require 'feidee_utils/record'

# A thin wrapper around SQLite3
module FeideeUtils
  class Database < SQLite3::Database
    Tables = {
      accounts: "t_account",
      account_groups: "t_account_group",
      categories: "t_category",
      transactions: "t_transaction",

      metadata: "t_metadata",
      profile: "t_profile",
      id_seed: "t_id_seed",
    }.freeze

    PotentialUsefulTables = %w(
      t_account_book
      t_account_info
      t_budget_item
      t_fund_holding
      t_fund_trans
      t_module_stock_holding
      t_module_stock_tran
      t_tradingEntity
      t_user
    ).freeze

    attr_reader :sqlite_file
    attr_reader :platform, :ledger_name, :last_modified_at
    attr_reader :missing_tables
    attr_reader :ledger

    def initialize(private_sqlite, strip = false)
      @sqlite_file = Database.feidee_to_sqlite(private_sqlite)

      super(@sqlite_file.path)

      extract_metadata
      drop_unused_tables if strip

      # TODO: make Ledger a first class object.
      @ledger = Record.generate_subclasses(self)
    end

    def sqlite_backup(dest_file_path)
      self.execute("vacuum;")

      backup_sqlite_db = SQLite3::Database.new(dest_file_path.to_s)
      backup_obj = SQLite3::Backup.new(backup_sqlite_db, "main", self, "main")
      backup_obj.step(-1)
      backup_obj.finish
      backup_sqlite_db.close
    end

    def validate_global_integrity
      @ledger.constants.each do |const|
        @ledger.const_get(const).validate_global_integrity if const != :Database
      end
    end

    private
    def all_tables
      rows = self.execute <<-SQL
        SELECT name FROM sqlite_master
        WHERE type IN ('table','view') AND name NOT LIKE 'sqlite_%'
        UNION ALL
        SELECT name FROM sqlite_temp_master
        WHERE type IN ('table','view')
        ORDER BY 1
      SQL
      rows.map do |row| row[0] end.sort
    end

    def drop_unused_tables
      useful_tables = Tables.values + PotentialUsefulTables
      useful_tables += useful_tables.map do |x|
        self.class.trash_table_name(x)
      end
      useful_tables.sort!

      # TODO: Record all tables droped.
      (all_tables - useful_tables).each do |table|
        self.execute("DROP TABLE IF EXISTS #{table}");
      end

      @missing_tables = Tables.values - all_tables
      if !@missing_tables.empty?
        raise "Missing tables: #{@missing_tables} from kbf backup."
      end
    end

    def extract_metadata
      @platform = self.execute(
        "SELECT platform from #{Tables[:metadata]}"
      )[0][0];

      @ledger_name = self.get_first_row(
        "SELECT accountBookName FROM #{Tables[:profile]};"
      )[0];

      # This is not recorded in the database, so the lastest lastUpdateTime of
      # all transactions is chosen.
      timestamp = self.get_first_row(
        "SELECT max(lastUpdateTime) FROM #{Tables[:transactions]};"
      )[0]
      @last_modified_at = Time.at((timestamp or 0.0) / 1000)
    end

    class << self
      def open_file(file_name)
        Database.new(File.open(file_name))
      end

      Header = "SQLite format 3\0".force_encoding("binary")
      FeideeHeader_iOS = "%$^#&!@_@- -!F\xff\0".force_encoding('binary')
      FeideeHeader_Android = ("\0" * 13 + "F\xff\0").force_encoding("binary")
      AllHeaders = [FeideeHeader_iOS, FeideeHeader_Android, Header]

      def feidee_to_sqlite(private_sqlite, sqlite_file = nil)
        # Discard the first a few bytes content.
        private_header = private_sqlite.read(Header.length)

        unless AllHeaders.include? private_header
          raise "Unexpected private sqlite header #{private_header.inspect}."
        end

        # Write the rest to a tempfile.
        sqlite_file ||= Tempfile.new("kingdee_sqlite", binmode: true)
        sqlite_file.binmode
        sqlite_file.write(Header)
        sqlite_file.write(private_sqlite.read)
        sqlite_file.fsync
        sqlite_file
      end

      def sqlite_to_feidee(sqlite, feidee_file = nil, platform = :iOS)
        sqlite.read(Header.length)
        private_header = if platform == :iOS
                           FeideeHeader_iOS
                         else
                           FeideeHeader_Android
                         end

        feidee_file ||= Tempfile.new("feidee_sqlite", binmode: true)
        feidee_file.binmode
        feidee_file.write(private_header)
        feidee_file.write(sqlite.read)
        feidee_file.fsync
        feidee_file
      end
    end

    class << self
      NoDeleteSuffixTables = %w(
        account category tag tradingEntity
        transaction transaction_template
      )

      def trash_table_name name
        NoDeleteSuffixTables.each do |core_name|
          if name == "t_" + core_name then
            return "t_" + "deleted_" + core_name;
          end
        end

        name + "_delete"
      end
    end

    AllKnownTables = {
      t_account:                "As named",
      t_account_book:           "A group of accounts, travel accounts etc.",
      t_account_extra:          "Extra Feidee account configs, key/value pair.",
      t_account_group:          "A group of accounts, saving/chekcing etc.",
      t_account_info:           "Additional info of accounts: banks etc.",
      t_accountgrant:           "Feidee account VIP related stuff.",
      t_budget_item:            "Used to create budgets. An extension of category.",
      t_binding:                "Netbank / credit card / Taobao bindings.",
      t_category:               "Transaction categories.",
      t_currency:               "Currency types.",
      t_exchange:               "Currency exchange rates.",
      t_fund:                   "List of money manage institute names. Abandoned.",
      t_fund_holding:           "Fund accounts.",
      t_fund_trans:             "Fund transactions.",
      t_fund_price_history:     "Fund price history",
      t_id_seed:                "ID seeds for all tables.",
      t_import_history:         "As named.",
      t_import_source:          "Import data from text messages etc.",

      # JCT stands for Jia Cai Tong, a software for family book keeping.
      # See http://www.feidee.com/jct/
      # JCT is quite obsolete. All the related tables can be safely ignored.
      t_jct_clientdeviceregist: "Client devices.",
      t_jct_clientdevicestatus: "Client devices status.",
      t_jct_syncbookfilelist:   "Name of files synced from other devices.",
      t_jct_usergrant:          "Maybe if the user has purchased any service.",
      t_jct_userlog:            "As named.",

      t_local_recent:           "Local merchandise used recently",
      t_message:                "Kingdee ads.",
      t_metadata:               "Database version, client version etc.",
      t_module_stock_holding:   "Stock accounts.",
      t_module_stock_info:      "Stock rates.",
      t_module_stock_trans:     "Stock transactions.",
      t_notification:           "If and when a notification has been delivered.",
      t_profile:                "User profile, default stuff when configuring account books.",
      t_property:               "Data collected on user settings, key/value pair.",
      t_syncResource:           "???",
      t_sync_logs:              "As named.",
      t_tag:                    "Other support like roles/merchandise.",
      t_tradingEntity:          "Merchandise. Used together with t_user in Debt Center.",
      t_trans_debt:             "Transactions in Debt Center.",
      t_trans_debt_group:       "Transaction groups in Debt Center.",
      t_transaction:            "As named.",
      t_transaction_projectcategory_map: "Transaction project.",
      t_transaction_template:   "As named. UI.",
      t_user:                   "Multi-user support.",
      t_usage_count:            "As named. Abandoned."
    }
  end
end
